'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import time
import uuid

from pm4py.objects.petri_net.obj import Marking
from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils.petri_utils import (
    remove_transition,
    add_arc_from_to,
    remove_place,
)
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.process_tree.obj import Operator
from pm4py.objects.petri_net.utils import reduction


class Counts(object):
    """
    Shared variables among executions
    """

    def __init__(self):
        """
        Constructor
        """
        self.num_places = 0
        self.num_hidden = 0
        self.num_visible_trans = 0
        self.dict_skips = {}
        self.dict_loops = {}

    def inc_places(self):
        """
        Increase the number of places
        """
        self.num_places = self.num_places + 1

    def inc_no_hidden(self):
        """
        Increase the number of hidden transitions
        """
        self.num_hidden = self.num_hidden + 1

    def inc_no_visible(self):
        """
        Increase the number of visible transitions
        """
        self.num_visible_trans = self.num_visible_trans + 1


def clean_duplicate_transitions(net):
    """
    Clean duplicate transitions in a Petri net

    Parameters
    ------------
    net
        Petri net

    Returns
    ------------
    net
        Cleaned Petri net
    """
    transitions = list(net.transitions)
    already_visited_combo = set()
    for i in range(0, len(transitions)):
        trans = transitions[i]
        if trans.label is None:
            in_arcs = trans.in_arcs
            out_arcs = trans.out_arcs
            to_delete = False
            for in_arc in in_arcs:
                in_place = in_arc.source
                for out_arc in out_arcs:
                    out_place = out_arc.target
                    combo = in_place.name + " " + out_place.name
                    if combo in already_visited_combo:
                        to_delete = True
                        break
                    already_visited_combo.add(combo)
            if to_delete:
                net = remove_transition(net, trans)
    return net


def get_new_place(counts):
    """
    Create a new place in the Petri net
    """
    counts.inc_places()
    return PetriNet.Place("p_" + str(counts.num_places))


def get_new_hidden_trans(counts, type_trans="unknown"):
    """
    Create a new hidden transition in the Petri net
    """
    counts.inc_no_hidden()
    return PetriNet.Transition(type_trans + "_" + str(counts.num_hidden), None)


def get_transition(counts, label):
    """
    Create a transitions with the specified label in the Petri net
    """
    counts.inc_no_visible()
    return PetriNet.Transition(str(uuid.uuid4()), label)


def get_first_terminal_child_transitions(tree):
    """
    Gets the list of transitions belonging to the first terminal child node of the current tree

    Parameters
    ----------
    tree
        Process tree

    Returns
    ---------
    transitions_list
        List of transitions belonging to the first terminal child node
    """
    if tree.children:
        if tree.children[0].operator:
            return get_first_terminal_child_transitions(tree.children[0])
        else:
            if tree.children[0].children:
                return tree.children[0].children
            else:
                return [tree.children[0]]
    return []


def get_last_terminal_child_transitions(tree):
    """
    Gets the list of transitions belonging to the last terminal child node of the current tree

    Parameters
    ----------
    tree
        Process tree

    Returns
    ---------
    transitions_list
        List of transitions belonging to the first terminal child node
    """
    if tree.children:
        if tree.children[-1].operator:
            return get_last_terminal_child_transitions(tree.children[-1])
        else:
            if tree.children[-1].children:
                return tree.children[-1].children
            else:
                return [tree.children[-1]]
    return []


def check_loop_to_first_operator(tree):
    """
    Checks if loop to first operator

    Parameters
    ------------
    tree
        Process tree

    Returns
    ------------
    boolean
        Check if no loop to the first operator
    """
    if tree.operator == Operator.LOOP:
        return True
    if tree.children:
        if tree.children[0].operator == Operator.LOOP:
            return True
        else:
            return check_loop_to_first_operator(tree.children[0])
    return tree.operator == Operator.LOOP


def check_loop_to_last_operator(tree):
    """
    Checks if loop to last operator

    Parameters
    -------------
    tree
        Process tree

    Returns
    -------------
    boolean
        Check if no loop to the last operator
    """
    if tree.operator == Operator.LOOP:
        return True
    if tree.children:
        if tree.children[-1].operator == Operator.LOOP:
            return True
        else:
            return check_loop_to_last_operator(tree.children[-1])
    return tree.operator == Operator.LOOP


def check_initial_loop(tree):
    """
    Check if the tree, on-the-left, starts with a loop

    Parameters
    ----------
    tree
        Process tree

    Returns
    ----------
    boolean
        True if it starts with an initial loop
    """
    if tree.children:
        if tree.children[0].operator:
            if tree.children[0].operator == Operator.LOOP:
                return True
            else:
                return check_terminal_loop(tree.children[0])
    return False


def check_terminal_loop(tree):
    """
    Check if the tree, on-the-right, ends with a loop

    Parameters
    ----------
    tree
        Process tree

    Returns
    -----------
    boolean
        True if it ends with a terminal loop
    """
    if tree.children:
        if tree.children[-1].operator:
            if tree.children[-1].operator == Operator.LOOP:
                return True
            else:
                return check_terminal_loop(tree.children[-1])
    return False


def check_tau_mandatory_at_initial_marking(tree):
    """
    When a conversion to a Petri net is operated, check if is mandatory to add a hidden transition
    at initial marking

    Parameters
    ----------
    tree
        Process tree

    Returns
    ----------
    boolean
        Boolean that is true if it is mandatory to add a hidden transition connecting the initial marking
        to the rest of the process
    """
    condition1 = check_initial_loop(tree)
    terminal_transitions = get_first_terminal_child_transitions(tree)
    condition2 = len(terminal_transitions) > 1
    condition3 = check_loop_to_first_operator(tree)
    condition4 = (
        tree.operator == Operator.XOR or tree.operator == Operator.PARALLEL
    )

    return condition1 or condition2 or condition3 or condition4


def check_tau_mandatory_at_final_marking(tree):
    """
    When a conversion to a Petri net is operated, check if is mandatory to add a hidden transition
    at final marking

    Returns
    ----------
    boolean
        Boolean that is true if it is mandatory to add a hidden transition connecting
        the rest of the process to the final marking
    """
    condition1 = check_terminal_loop(tree)
    terminal_transitions = get_last_terminal_child_transitions(tree)
    condition2 = len(terminal_transitions) > 1
    condition3 = check_loop_to_last_operator(tree)
    condition4 = (
        tree.operator == Operator.XOR or tree.operator == Operator.PARALLEL
    )

    return condition1 or condition2 or condition3 or condition4


def recursively_add_tree(
    parent_tree,
    tree,
    net,
    initial_entity_subtree,
    final_entity_subtree,
    counts,
    rec_depth,
    force_add_skip=False,
):
    """
    Recursively add the subtrees to the Petri net

    Parameters
    -----------
    parent_tree
        Parent tree
    tree
        Current subtree
    net
        Petri net
    initial_entity_subtree
        Initial entity (place/transition) that should be attached from the subtree
    final_entity_subtree
        Final entity (place/transition) that should be attached from the subtree
    counts
        Counts object (keeps the number of places, transitions and hidden transitions)
    rec_depth
        Recursion depth of the current iteration
    force_add_skip
        Boolean value that tells if the addition of a skip is mandatory

    Returns
    ----------
    net
        Updated Petri net
    counts
        Updated counts object (keeps the number of places, transitions and hidden transitions)
    final_place
        Last place added in this recursion
    """
    if type(initial_entity_subtree) is PetriNet.Transition:
        initial_place = get_new_place(counts)
        net.places.add(initial_place)
        add_arc_from_to(initial_entity_subtree, initial_place, net)
    else:
        initial_place = initial_entity_subtree
    if (
        final_entity_subtree is not None
        and type(final_entity_subtree) is PetriNet.Place
    ):
        final_place = final_entity_subtree
    else:
        final_place = get_new_place(counts)
        net.places.add(final_place)
        if (
            final_entity_subtree is not None
            and type(final_entity_subtree) is PetriNet.Transition
        ):
            add_arc_from_to(final_place, final_entity_subtree, net)
    tree_childs = [child for child in tree.children]

    if force_add_skip:
        invisible = get_new_hidden_trans(counts, type_trans="skip")
        add_arc_from_to(initial_place, invisible, net)
        add_arc_from_to(invisible, final_place, net)

    if tree.operator is None:
        trans = tree
        if trans.label is None:
            petri_trans = get_new_hidden_trans(counts, type_trans="skip")
        else:
            petri_trans = get_transition(counts, trans.label)
        net.transitions.add(petri_trans)
        add_arc_from_to(initial_place, petri_trans, net)
        add_arc_from_to(petri_trans, final_place, net)

    if tree.operator == Operator.XOR:
        for subtree in tree_childs:
            net, counts, intermediate_place = recursively_add_tree(
                tree,
                subtree,
                net,
                initial_place,
                final_place,
                counts,
                rec_depth + 1,
            )
    elif tree.operator == Operator.OR:
        new_initial_trans = get_new_hidden_trans(counts, type_trans="tauSplit")
        net.transitions.add(new_initial_trans)
        add_arc_from_to(initial_place, new_initial_trans, net)
        new_final_trans = get_new_hidden_trans(counts, type_trans="tauJoin")
        net.transitions.add(new_final_trans)
        add_arc_from_to(new_final_trans, final_place, net)
        terminal_place = get_new_place(counts)
        net.places.add(terminal_place)
        add_arc_from_to(terminal_place, new_final_trans, net)
        first_place = get_new_place(counts)
        net.places.add(first_place)
        add_arc_from_to(new_initial_trans, first_place, net)

        for subtree in tree_childs:
            subtree_init_place = get_new_place(counts)
            net.places.add(subtree_init_place)
            add_arc_from_to(new_initial_trans, subtree_init_place, net)
            subtree_start_place = get_new_place(counts)
            net.places.add(subtree_start_place)
            subtree_end_place = get_new_place(counts)
            net.places.add(subtree_end_place)
            trans_start = get_new_hidden_trans(
                counts, type_trans="inclusiveStart"
            )
            trans_later = get_new_hidden_trans(
                counts, type_trans="inclusiveLater"
            )
            trans_skip = get_new_hidden_trans(
                counts, type_trans="inclusiveSkip"
            )
            net.transitions.add(trans_start)
            net.transitions.add(trans_later)
            net.transitions.add(trans_skip)
            add_arc_from_to(first_place, trans_start, net)
            add_arc_from_to(subtree_init_place, trans_start, net)
            add_arc_from_to(trans_start, subtree_start_place, net)
            add_arc_from_to(trans_start, terminal_place, net)

            add_arc_from_to(terminal_place, trans_later, net)
            add_arc_from_to(subtree_init_place, trans_later, net)
            add_arc_from_to(trans_later, subtree_start_place, net)
            add_arc_from_to(trans_later, terminal_place, net)

            add_arc_from_to(terminal_place, trans_skip, net)
            add_arc_from_to(subtree_init_place, trans_skip, net)
            add_arc_from_to(trans_skip, terminal_place, net)
            add_arc_from_to(trans_skip, subtree_end_place, net)

            add_arc_from_to(subtree_end_place, new_final_trans, net)

            net, counts, intermediate_place = recursively_add_tree(
                tree,
                subtree,
                net,
                subtree_start_place,
                subtree_end_place,
                counts,
                rec_depth + 1,
            )

    elif tree.operator == Operator.PARALLEL:
        new_initial_trans = get_new_hidden_trans(counts, type_trans="tauSplit")
        net.transitions.add(new_initial_trans)
        add_arc_from_to(initial_place, new_initial_trans, net)
        new_final_trans = get_new_hidden_trans(counts, type_trans="tauJoin")
        net.transitions.add(new_final_trans)
        add_arc_from_to(new_final_trans, final_place, net)

        for subtree in tree_childs:
            net, counts, intermediate_place = recursively_add_tree(
                tree,
                subtree,
                net,
                new_initial_trans,
                new_final_trans,
                counts,
                rec_depth + 1,
            )

    elif tree.operator == Operator.INTERLEAVING:
        new_initial_trans = get_new_hidden_trans(counts, type_trans="tauSplit")
        net.transitions.add(new_initial_trans)
        add_arc_from_to(initial_place, new_initial_trans, net)
        new_final_trans = get_new_hidden_trans(counts, type_trans="tauJoin")
        net.transitions.add(new_final_trans)
        add_arc_from_to(new_final_trans, final_place, net)

        control_place = get_new_place(counts)
        net.places.add(control_place)

        add_arc_from_to(new_initial_trans, control_place, net)
        add_arc_from_to(control_place, new_final_trans, net)

        for subtree in tree_childs:
            placeI = get_new_place(counts)
            net.places.add(placeI)
            iTrans = get_new_hidden_trans(counts, type_trans="iTrans")
            net.transitions.add(iTrans)
            placeF = get_new_place(counts)
            net.places.add(placeF)
            fTrans = get_new_hidden_trans(counts, type_trans="fTrans")
            net.transitions.add(fTrans)

            add_arc_from_to(new_initial_trans, placeI, net)
            add_arc_from_to(placeI, iTrans, net)
            add_arc_from_to(fTrans, placeF, net)
            add_arc_from_to(placeF, new_final_trans, net)

            add_arc_from_to(control_place, iTrans, net)
            add_arc_from_to(fTrans, control_place, net)

            net, counts, intermediate_place = recursively_add_tree(
                tree, subtree, net, iTrans, fTrans, counts, rec_depth + 1
            )

    elif tree.operator == Operator.SEQUENCE:
        intermediate_place = initial_place
        for i in range(len(tree_childs)):
            final_connection_place = None
            if i == len(tree_childs) - 1:
                final_connection_place = final_place
            net, counts, intermediate_place = recursively_add_tree(
                tree,
                tree_childs[i],
                net,
                intermediate_place,
                final_connection_place,
                counts,
                rec_depth + 1,
            )
    elif tree.operator == Operator.LOOP:
        # if not parent_tree.operator == Operator.SEQUENCE:
        new_initial_place = get_new_place(counts)
        net.places.add(new_initial_place)
        init_loop_trans = get_new_hidden_trans(counts, type_trans="init_loop")
        net.transitions.add(init_loop_trans)
        add_arc_from_to(initial_place, init_loop_trans, net)
        add_arc_from_to(init_loop_trans, new_initial_place, net)
        initial_place = new_initial_place
        loop_trans = get_new_hidden_trans(counts, type_trans="loop")
        net.transitions.add(loop_trans)
        if len(tree_childs) == 1:
            net, counts, intermediate_place = recursively_add_tree(
                tree,
                tree_childs[0],
                net,
                initial_place,
                final_place,
                counts,
                rec_depth + 1,
            )
            add_arc_from_to(final_place, loop_trans, net)
            add_arc_from_to(loop_trans, initial_place, net)
        else:
            net, counts, int1 = recursively_add_tree(
                tree,
                tree_childs[0],
                net,
                initial_place,
                None,
                counts,
                rec_depth + 1,
            )
            int2 = None
            for i in range(1, len(tree_childs)):
                net, counts, int2 = recursively_add_tree(
                    tree,
                    tree_childs[i],
                    net,
                    int1,
                    int2,
                    counts,
                    rec_depth + 1,
                )

            net, counts, int3 = recursively_add_tree(
                tree,
                ProcessTree(),
                net,
                int1,
                final_place,
                counts,
                rec_depth + 1,
            )

            looping_place = int2

            add_arc_from_to(looping_place, loop_trans, net)
            add_arc_from_to(loop_trans, initial_place, net)

    return net, counts, final_place


def apply(tree, parameters=None):
    """
    Apply from Process Tree to Petri net

    Parameters
    -----------
    tree
        Process tree
    parameters
        Parameters of the algorithm

    Returns
    -----------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """
    if parameters is None:
        parameters = {}
    del parameters

    counts = Counts()
    net = PetriNet("imdf_net_" + str(time.time()))
    initial_marking = Marking()
    final_marking = Marking()
    source = get_new_place(counts)
    source.name = "source"
    sink = get_new_place(counts)
    sink.name = "sink"
    net.places.add(source)
    net.places.add(sink)
    initial_marking[source] = 1
    final_marking[sink] = 1
    initial_mandatory = check_tau_mandatory_at_initial_marking(tree)
    final_mandatory = check_tau_mandatory_at_final_marking(tree)
    if initial_mandatory:
        initial_place = get_new_place(counts)
        net.places.add(initial_place)
        tau_initial = get_new_hidden_trans(counts, type_trans="tau")
        net.transitions.add(tau_initial)
        add_arc_from_to(source, tau_initial, net)
        add_arc_from_to(tau_initial, initial_place, net)
    else:
        initial_place = source
    if final_mandatory:
        final_place = get_new_place(counts)
        net.places.add(final_place)
        tau_final = get_new_hidden_trans(counts, type_trans="tau")
        net.transitions.add(tau_final)
        add_arc_from_to(final_place, tau_final, net)
        add_arc_from_to(tau_final, sink, net)
    else:
        final_place = sink

    net, counts, last_added_place = recursively_add_tree(
        tree, tree, net, initial_place, final_place, counts, 0
    )

    reduction.apply_simple_reduction(net)

    places = list(net.places)
    for place in places:
        if len(place.out_arcs) == 0 and place not in final_marking:
            remove_place(net, place)
        if len(place.in_arcs) == 0 and place not in initial_marking:
            remove_place(net, place)

    return net, initial_marking, final_marking



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys


def indent_representation(model_string, max_indent=sys.maxsize):
    model_string = (
        model_string.replace("\n", "")
        .replace("\r", "")
        .replace("\t", "")
        .strip()
    )

    indent_level = 0
    list_strs = []
    formatted_str = ""
    for char in model_string:
        if char in "({":
            formatted_str += char
            indent_level += 1
            if indent_level <= max_indent:
                list_strs.append(formatted_str)
                formatted_str = "\t" * indent_level
        elif char in ")}":
            indent_level -= 1
            if indent_level < max_indent:
                if formatted_str[-1] not in "({":
                    list_strs.append(formatted_str)
                    formatted_str = "\t" * indent_level
            formatted_str += char
        elif char == ",":
            formatted_str += char
            if indent_level <= max_indent:
                list_strs.append(formatted_str)
                formatted_str = "\t" * indent_level
        else:
            formatted_str += char
    list_strs.append(formatted_str)
    return list_strs



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.algo.discovery.ocel.ocpn.variants import classic
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import exec_utils
from typing import Optional, Dict, Any


class Variants(Enum):
    WO_ANNOTATION = classic
    CLASSIC = classic


def apply(
    ocel: OCEL,
    variant=Variants.CLASSIC,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Discovers an object-centric Petri net from the provided object-centric event log.

    Reference paper: van der Aalst, Wil MP, and Alessandro Berti. "Discovering object-centric Petri nets." Fundamenta informaticae 175.1-4 (2020): 1-40.

    Parameters
    -----------------
    ocel
        Object-centric event log
    variant
        Variant of the algorithm to be used
    parameters
        Variant-specific parameters

    Returns
    ----------------
    ocpn
        Object-centric Petri net model, as a dictionary of properties.
    """
    return exec_utils.get_variant(variant).apply(ocel, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from typing import Optional, Tuple, List, Any, Dict

from pm4py.algo.discovery.inductive.cuts.factory import CutFactory
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree, Operator
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL
from enum import Enum
from pm4py.util import exec_utils, constants


class Parameters(Enum):
    MULTIPROCESSING = "multiprocessing"


class ActivityConcurrentUVCL(FallThrough[IMDataStructureUVCL]):
    MULTI_PROCESSING_LOWER_BOUND = 20

    @classmethod
    def _process_candidate(
        cls,
        c: Any,
        log: UVCL,
        queue=None,
        ev=None,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        l_alt = Counter()
        for t in log:
            l_alt[tuple(filter(lambda e: e != c, t))] = log[t]
        cut = cls._find_cut(
            IMDataStructureUVCL(l_alt), ev, parameters=parameters
        )
        if queue is not None:
            queue.put((c, cut))
        return cut if cut is not None else None

    @classmethod
    def _get_candidate(
        cls,
        obj: IMDataStructureUVCL,
        pool,
        manager,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Any]:
        if parameters is None:
            parameters = {}

        enable_multiprocessing = exec_utils.get_param_value(
            Parameters.MULTIPROCESSING,
            parameters,
            constants.ENABLE_MULTIPROCESSING_DEFAULT,
        )

        log = obj.data_structure
        candidates = sorted(list(comut.get_alphabet(log)))
        if (
            pool is None
            or manager is None
            or not enable_multiprocessing
            or len(candidates)
            <= ActivityConcurrentUVCL.MULTI_PROCESSING_LOWER_BOUND
        ):
            for a in candidates:
                cut = cls._process_candidate(a, log, parameters=parameters)
                if cut is not None:
                    return a
        else:
            q = manager.Queue()
            ev = manager.Event()
            # avoid dangerous freealloc from Python's garbage collector
            manager.support_list.append(q)
            manager.support_list.append(ev)

            for a in candidates:
                pool.apply_async(
                    cls._process_candidate, (a, log, q, ev, parameters)
                )
            potentials = set(candidates)
            while len(potentials) > 0:
                (c, cut) = q.get(block=True)
                if cut is None:
                    potentials.remove(c)
                else:
                    ev.set()
                    return c

        return None

    @classmethod
    def _find_cut(
        cls,
        obj: IMDataStructureUVCL,
        ev,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[IMDataStructureUVCL]]]:
        for c in CutFactory.get_cuts(
            obj, IMInstance.IM, parameters=parameters
        ):
            if ev is not None and ev.is_set():
                return None
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None

    @classmethod
    def holds(
        cls,
        obj: IMDataStructureUVCL,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> bool:
        return cls._get_candidate(obj, None, None, parameters) is not None

    @classmethod
    def apply(
        cls,
        obj: IMDataStructureUVCL,
        pool=None,
        manager=None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[IMDataStructureUVCL]]]:
        candidate = cls._get_candidate(obj, pool, manager, parameters)
        if candidate is None:
            return None
        log = obj.data_structure
        l_a = Counter()
        l_other = Counter()
        for t in log:
            l_a.update({tuple(filter(lambda e: e == candidate, t)): log[t]})
            l_other.update(
                {tuple(filter(lambda e: e != candidate, t)): log[t]}
            )
        return ProcessTree(operator=Operator.PARALLEL), [
            IMDataStructureUVCL(l_a),
            IMDataStructureUVCL(l_other),
        ]



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import polars as pl
from typing import Optional, Dict, Any
from pm4py.util import constants, xes_constants, exec_utils
from enum import Enum


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    FILTER_TYPE = "filter_type"


def apply(
    df: pl.LazyFrame,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pl.LazyFrame:
    """
    Groups the consecutive events of the same case having the same activity, providing option to keep the first/last event of each group

    Parameters
    ---------------
    df
        Polars LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY => the case identifier to be used
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.FILTER_TYPE => the type of filter to be applied:
            first => keeps the first event of each group
            last => keeps the last event of each group

    Returns
    ---------------
    filtered_dataframe
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    filter_type = exec_utils.get_param_value(
        Parameters.FILTER_TYPE, parameters, "first"
    )

    # Add row number to maintain order
    df = df.with_columns(pl.lit(1).cum_sum().alias("_row_num"))
    
    # Create a column that changes when case or activity changes
    df = df.with_columns(
        (
            (pl.col(case_id_key) != pl.col(case_id_key).shift(1)) |
            (pl.col(activity_key) != pl.col(activity_key).shift(1))
        ).fill_null(True).cum_sum().alias("_group_id")
    )

    if filter_type == "first":
        # Keep first event of each group
        df = df.group_by("_group_id").agg(pl.all().first()).sort("_row_num")
    elif filter_type == "last":
        # Keep last event of each group
        df = df.group_by("_group_id").agg(pl.all().last()).sort("_row_num")

    # Remove helper columns
    df = df.drop(["_row_num", "_group_id"])

    return df


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.objects.ocel import constants
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE


def get(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    if parameters is None:
        parameters = {}

    event_activity_column = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, ocel.event_activity
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )

    ets = {
        k: {
            x: str(v[x].dtype)
            for x in v.dropna(axis="columns", how="all").columns
            if not x.startswith("ocel:")
        }
        for k, v in ocel.events.groupby(event_activity_column)
    }
    ots = {
        k: {
            x: str(v[x].dtype)
            for x in v.dropna(axis="columns", how="all").columns
            if not x.startswith("ocel:")
        }
        for k, v in ocel.objects.groupby(object_type_column)
    }
    ots2 = {
        k: {
            x: str(v[x].dtype)
            for x in v.dropna(axis="columns", how="all").columns
            if not x.startswith("ocel:")
        }
        for k, v in ocel.object_changes.groupby(object_type_column)
    }

    for k in ots2:
        if k not in ots:
            ots[k] = ots2[k]
        else:
            for x in ots2[k]:
                if x not in ots[k]:
                    ots[k][x] = ots2[k][x]

    return ets, ots



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Tuple, List, TypeVar, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureLog
from pm4py.algo.discovery.powl.inductive.variants.dynamic_clustering_frequency.factory import (
    CutFactoryPOWLDynamicClusteringFrequency, )
from pm4py.algo.discovery.powl.inductive.variants.im_tree import IMBasePOWL
from pm4py.algo.discovery.powl.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant, )
from pm4py.objects.powl.obj import POWL

T = TypeVar("T", bound=IMDataStructureLog)


class POWLInductiveMinerDynamicClusteringFrequency(IMBasePOWL):

    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DYNAMIC_CLUSTERING

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLDynamicClusteringFrequency.find_cut(
            obj, parameters=parameters
        )
        return res



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.footprints.log.variants import (
    entire_event_log,
    trace_by_trace,
    entire_dataframe,
)
from pm4py.algo.discovery.footprints.petri.variants import reach_graph
from pm4py.algo.discovery.footprints.dfg.variants import dfg
from pm4py.algo.discovery.footprints.tree.variants import bottomup
from pm4py.algo.discovery.footprints.powl.variants import bottomup as bottomup_powl
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.process_tree.obj import ProcessTree
from enum import Enum
from pm4py.util import exec_utils, pandas_utils
from pm4py.utils import is_polars_lazyframe
import pandas as pd
from typing import Optional, Dict, Any
import importlib.util

POLARS_AVAILABLE = importlib.util.find_spec("polars") is not None
if POLARS_AVAILABLE:
    from pm4py.algo.discovery.footprints.log.variants import polars_lazyframes


class Variants(Enum):
    ENTIRE_EVENT_LOG = entire_event_log
    ENTIRE_DATAFRAME = entire_dataframe
    TRACE_BY_TRACE = trace_by_trace
    PETRI_REACH_GRAPH = reach_graph
    PROCESS_TREE = bottomup
    POWL = bottomup_powl
    DFG = dfg
    POLARS_LAZYFRAMES = polars_lazyframes if POLARS_AVAILABLE else entire_dataframe


def apply(
    *args, variant=None, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, Any]:
    """
    Discovers a footprint object from a log/model

    Parameters
    --------------
    args
        Positional arguments that describe the log/model
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm, including:
            - Variants.ENTIRE_EVENT_LOG
            - Variants.TRACE_BY_TRACE
            - Variants.PETRI_REACH_GRAPH
            - Variants.DFG

    Returns
    --------------
    footprints_obj
        Footprints object
    """
    from pm4py.objects.powl.obj import POWL, StrictPartialOrder, OperatorPOWL

    if variant is None:
        if type(args[0]) is EventLog:
            variant = Variants.TRACE_BY_TRACE
        elif type(args[0]) is PetriNet:
            variant = Variants.PETRI_REACH_GRAPH
        elif isinstance(args[0], POWL):
            variant = Variants.POWL
        elif type(args[0]) is ProcessTree:
            variant = Variants.PROCESS_TREE
        elif isinstance(args[0], dict):
            variant = Variants.DFG

        if pandas_utils.check_is_pandas_dataframe(args[0]):
            if is_polars_lazyframe(args[0]):
                if not POLARS_AVAILABLE:
                    raise RuntimeError(
                        "Polars LazyFrame provided but 'polars' package is not installed."
                    )
                variant = Variants.POLARS_LAZYFRAMES
            else:
                variant = Variants.ENTIRE_DATAFRAME

        if variant is None:
            return Exception("unsupported arguments")

    if variant in [
        Variants.TRACE_BY_TRACE,
        Variants.ENTIRE_EVENT_LOG,
        Variants.DFG,
        Variants.PROCESS_TREE,
        Variants.POWL,
        Variants.ENTIRE_DATAFRAME,
        Variants.POLARS_LAZYFRAMES,
    ]:
        return exec_utils.get_variant(variant).apply(
            args[0], parameters=parameters
        )
    elif variant in [Variants.PETRI_REACH_GRAPH]:
        return exec_utils.get_variant(variant).apply(
            args[0], args[1], parameters=parameters
        )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import time

from pm4py import (
    util,
    objects,
    statistics,
    algo,
    visualization,
    llm,
    connectors,
)
from pm4py import (
    analysis,
    conformance,
    convert,
    discovery,
    filtering,
    hof,
    ml,
    ocel,
    org,
    read,
    sim,
    stats,
    utils,
    vis,
    write,
)
from pm4py.read import (
    read_xes,
    read_dfg,
    read_bpmn,
    read_pnml,
    read_ptml,
    read_ocel,
    read_ocel_csv,
    read_ocel_xml,
    read_ocel_json,
    read_ocel_sqlite,
    read_ocel2,
    read_ocel2_sqlite,
    read_ocel2_json,
    read_ocel2_xml,
)
from pm4py.write import (
    write_xes,
    write_dfg,
    write_bpmn,
    write_pnml,
    write_ptml,
    write_ocel,
    write_ocel_json,
    write_ocel_csv,
    write_ocel_xml,
    write_ocel_sqlite,
    write_ocel2,
    write_ocel2_sqlite,
    write_ocel2_xml,
    write_ocel2_json,
)
from pm4py.utils import (
    format_dataframe,
    parse_process_tree,
    serialize,
    deserialize,
    set_classifier,
    parse_event_log_string,
    project_on_event_attribute,
    sample_cases,
    sample_events,
    rebase,
    parse_powl_model_string,
)
from pm4py.filtering import (
    filter_log_relative_occurrence_event_attribute,
    filter_start_activities,
    filter_end_activities,
    filter_variants,
    filter_directly_follows_relation,
    filter_time_range,
    filter_eventually_follows_relation,
    filter_event_attribute_values,
    filter_trace_attribute_values,
    filter_between,
    filter_case_size,
    filter_case_performance,
    filter_activities_rework,
    filter_paths_performance,
    filter_variants_by_coverage_percentage,
    filter_variants_top_k,
    filter_ocel_event_attribute,
    filter_ocel_object_attribute,
    filter_ocel_object_types_allowed_activities,
    filter_ocel_object_per_type_count,
    filter_ocel_start_events_per_object_type,
    filter_ocel_end_events_per_object_type,
    filter_ocel_events_timestamp,
    filter_prefixes,
    filter_suffixes,
    filter_trace_segments,
    filter_four_eyes_principle,
    filter_activity_done_different_resources,
    filter_ocel_events,
    filter_ocel_objects,
    filter_ocel_object_types,
    filter_ocel_cc_object,
    filter_ocel_cc_length,
    filter_ocel_cc_otype,
    filter_ocel_cc_activity,
    filter_ocel_activities_connected_object_type,
    filter_dfg_activities_percentage,
    filter_dfg_paths_percentage
)
from pm4py.discovery import (
    discover_petri_net_alpha,
    discover_petri_net_alpha_plus,
    discover_petri_net_ilp,
    discover_petri_net_heuristics,
    discover_petri_net_inductive,
    discover_process_tree_inductive,
    discover_heuristics_net,
    discover_dfg,
    discover_footprints,
    discover_eventually_follows_graph,
    discover_directly_follows_graph,
    discover_bpmn_inductive,
    discover_performance_dfg,
    discover_transition_system,
    discover_prefix_tree,
    discover_temporal_profile,
    discover_log_skeleton,
    discover_batches,
    derive_minimum_self_distance,
    discover_dfg_typed,
    discover_declare,
    discover_powl,
    correlation_miner,
    discover_otg,
    discover_etot,
)
from pm4py.conformance import (
    conformance_diagnostics_token_based_replay,
    conformance_diagnostics_alignments,
    fitness_token_based_replay,
    fitness_alignments,
    precision_token_based_replay,
    precision_alignments,
    conformance_diagnostics_footprints,
    fitness_footprints,
    precision_footprints,
    check_is_fitting,
    conformance_temporal_profile,
    conformance_declare,
    conformance_log_skeleton,
    conformance_ocdfg,
    conformance_otg,
    conformance_etot,
    replay_prefix_tbr,
    generalization_tbr,
)
from pm4py.ocel import (
    ocel_objects_interactions_summary,
    ocel_temporal_summary,
    ocel_objects_summary,
    ocel_get_object_types,
    ocel_get_attribute_names,
    ocel_flattening,
    ocel_object_type_activities,
    ocel_objects_ot_count,
    discover_ocdfg,
    discover_oc_petri_net,
    discover_objects_graph,
    sample_ocel_objects,
    ocel_drop_duplicates,
    ocel_merge_duplicates,
    ocel_sort_by_additional_column,
    ocel_add_index_based_timedelta,
    sample_ocel_connected_components,
    ocel_o2o_enrichment,
    ocel_e2o_lifecycle_enrichment,
    cluster_equivalent_ocel,
)
from pm4py.vis import (
    view_petri_net,
    save_vis_petri_net,
    view_dfg,
    save_vis_dfg,
    view_process_tree,
    save_vis_process_tree,
    view_ocdfg,
    save_vis_ocdfg,
    view_heuristics_net,
    save_vis_heuristics_net,
    view_bpmn,
    save_vis_bpmn,
    view_sna,
    save_vis_sna,
    view_dotted_chart,
    save_vis_dotted_chart,
    view_performance_spectrum,
    save_vis_performance_spectrum,
    view_case_duration_graph,
    view_events_per_time_graph,
    save_vis_case_duration_graph,
    save_vis_events_per_time_graph,
    view_events_distribution_graph,
    save_vis_events_distribution_graph,
    view_performance_dfg,
    save_vis_performance_dfg,
    view_ocpn,
    save_vis_ocpn,
    view_network_analysis,
    save_vis_network_analysis,
    view_transition_system,
    save_vis_transition_system,
    view_prefix_tree,
    save_vis_prefix_tree,
    view_object_graph,
    save_vis_object_graph,
    view_alignments,
    save_vis_alignments,
    view_footprints,
    save_vis_footprints,
    view_powl,
    save_vis_powl,
)
from pm4py.convert import (
    convert_to_event_log,
    convert_to_event_stream,
    convert_to_dataframe,
    convert_to_bpmn,
    convert_to_petri_net,
    convert_to_process_tree,
    convert_to_reachability_graph,
    convert_log_to_ocel,
    convert_ocel_to_networkx,
    convert_log_to_networkx,
    convert_log_to_time_intervals,
    convert_petri_net_to_networkx,
    convert_petri_net_type,
    convert_to_powl,
)
from pm4py.analysis import (
    cluster_log,
    check_soundness,
    compute_emd,
    solve_marking_equation,
    solve_extended_marking_equation,
    construct_synchronous_product_net,
    insert_artificial_start_end,
    check_is_workflow_net,
    maximal_decomposition,
    generate_marking,
    reduce_petri_net_invisibles,
    reduce_petri_net_implicit_places,
    insert_case_arrival_finish_rate,
    insert_case_service_waiting_time,
    get_enabled_transitions,
    simplicity_petri_net,
    behavioral_similarity,
    structural_similarity,
    embeddings_similarity,
    get_activity_labels,
    replace_activity_labels,
    label_sets_similarity,
    map_labels_from_second_model
)
from pm4py.stats import (
    get_start_activities,
    get_process_cube,
    get_end_activities,
    get_event_attributes,
    get_event_attribute_values,
    get_variants,
    get_trace_attributes,
    get_variants_as_tuples,
    get_trace_attribute_values,
    get_case_arrival_average,
    get_minimum_self_distances,
    get_minimum_self_distance_witnesses,
    get_frequent_trace_segments,
    get_case_arrival_average,
    get_rework_cases_per_activity,
    get_case_overlap,
    get_cycle_time,
    get_all_case_durations,
    get_case_duration,
    get_activity_position_summary,
    get_stochastic_language,
    split_by_process_variant,
    get_variants_paths_duration,
    get_service_time
)
from pm4py.sim import play_out, generate_process_tree
from pm4py.ml import (
    split_train_test,
    get_prefixes_from_log,
    extract_ocel_features,
    extract_features_dataframe,
    extract_temporal_features_dataframe,
    extract_outcome_enriched_dataframe,
    extract_target_vector,
)
from pm4py.org import (
    discover_handover_of_work_network,
    discover_activity_based_resource_similarity,
    discover_subcontracting_network,
    discover_working_together_network,
    discover_organizational_roles,
    discover_network_analysis,
)

# from pm4py.hof import filter_log, filter_trace, sort_trace, sort_log
from pm4py.meta import (
    __name__,
    __version__,
    __doc__,
    __author__,
    __author_email__,
    __maintainer__,
    __maintainer_email__,
)

from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.bpmn.obj import BPMN

time.clock = time.process_time



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.conformance.log_skeleton.variants import classic
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List, Set
from pm4py.objects.log.obj import EventLog, Trace
import pandas as pd


class Variants(Enum):
    CLASSIC = classic


CLASSIC = Variants.CLASSIC
DEFAULT_VARIANT = Variants.CLASSIC


def apply(
    obj: Union[EventLog, Trace, pd.DataFrame],
    model: Dict[str, Any],
    variant=DEFAULT_VARIANT,
    parameters: Optional[Dict[Any, Any]] = None,
) -> List[Set[Any]]:
    """
    Apply log-skeleton based conformance checking given an event log/trace
    and a log-skeleton model

    Parameters
    --------------
    obj
        Object (event log/trace)
    model
        Log-skeleton model
    variant
        Variant of the algorithm, possible values: Variants.CLASSIC
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.CONSIDERED_CONSTRAINTS, among: equivalence, always_after, always_before, never_together, directly_follows, activ_freq

    Returns
    --------------
    aligned_traces
        Conformance checking results for each trace:
        - Outputs.IS_FIT => boolean that tells if the trace is perfectly fit according to the model
        - Outputs.DEV_FITNESS => deviation based fitness (between 0 and 1; the more the trace is near to 1 the more fit is)
        - Outputs.DEVIATIONS => list of deviations in the model
    """
    if parameters is None:
        parameters = {}

    if type(obj) is Trace:
        return exec_utils.get_variant(variant).apply_trace(
            obj, model, parameters=parameters
        )
    else:
        return exec_utils.get_variant(variant).apply_log(
            obj, model, parameters=parameters
        )


def apply_from_variants_list(
    var_list: List[List[str]],
    model: Dict[str, Any],
    variant=DEFAULT_VARIANT,
    parameters: Optional[Dict[Any, Any]] = None,
) -> List[Set[Any]]:
    """
    Performs conformance checking using the log skeleton,
    applying it from a list of variants

    Parameters
    --------------
    var_list
        List of variants
    model
        Log skeleton model
    variant
        Variant of the algorithm, possible values: Variants.CLASSIC
    parameters
        Parameters

    Returns
    --------------
    conformance_dictio
        Dictionary containing, for each variant, the result
        of log skeleton checking
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply_from_variants_list(
        var_list, model, parameters=parameters
    )


def get_diagnostics_dataframe(
    log: EventLog,
    conf_result: List[Set[Any]],
    variant=DEFAULT_VARIANT,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pd.DataFrame:
    """
    Gets the diagnostics dataframe from a log and the results
    of log skeleton-based conformance checking

    Parameters
    --------------
    log
        Event log
    conf_result
        Results of conformance checking

    Returns
    --------------
    diagn_dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).get_diagnostics_dataframe(
        log, conf_result, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import tempfile
from copy import copy
import sys

from graphviz import Digraph
from pm4py.util import constants
from typing import Dict, List, Tuple
from collections import defaultdict, deque
from pm4py.util.vis_utils import (
    human_readable_stat,
    get_arc_penwidth,
    get_trans_freq_color,
    value_to_color,
)


def get_activities_color(activities_count):
    """
    Get frequency color for attributes

    Parameters
    -----------
    activities_count
        Count of attributes in the log

    Returns
    -----------
    activities_color
        Color assigned to attributes in the graph
    """
    activities_color = {}

    min_value, max_value = get_min_max_value(activities_count)

    for ac in activities_count:
        v0 = activities_count[ac]
        v1 = get_trans_freq_color(v0, min_value, max_value)

        activities_color[ac] = v1

    return activities_color


def get_activities_color_serv_time(serv_time):
    """
    Gets the color for the activities based on the service time

    Parameters
    ----------------
    serv_time
        Service time

    Returns
    ----------------
    act_color
        Dictionary associating each activity to a color based on the service time
    """
    activities_color = {}

    min_soj_time, max_soj_time = get_min_max_value(serv_time)

    for ac in serv_time:
        act_soj_time = serv_time[ac]

        trans_base_color = int(
            255
            - 100
            * (act_soj_time - min_soj_time)
            / (max_soj_time - min_soj_time + 0.00001)
        )
        trans_base_color_hex = str(hex(trans_base_color))[2:].upper()

        activities_color[ac] = (
            "#" + "FF" + trans_base_color_hex + trans_base_color_hex
        )

    return activities_color


def get_min_max_value(dfg):
    """
    Gets min and max value assigned to edges
    in DFG graph

    Parameters
    -----------
    dfg
        Directly follows graph

    Returns
    -----------
    min_value
        Minimum value in directly follows graph
    max_value
        Maximum value in directly follows graph
    """
    min_value = 9999999999
    max_value = -1

    for edge in dfg:
        if dfg[edge] < min_value:
            min_value = dfg[edge]
        if dfg[edge] > max_value:
            max_value = dfg[edge]

    return min_value, max_value


def assign_penwidth_edges(dfg):
    """
    Assign penwidth to edges in directly-follows graph

    Parameters
    -----------
    dfg
        Direcly follows graph

    Returns
    -----------
    penwidth
        Graph penwidth that edges should have in the direcly follows graph
    """
    penwidth = {}
    min_value, max_value = get_min_max_value(dfg)
    for edge in dfg:
        v0 = dfg[edge]
        v1 = get_arc_penwidth(v0, min_value, max_value)
        penwidth[edge] = str(v1)

    return penwidth


def sort_dfg_reachability(
    dfg: List[Tuple[str, str]],
    start_activities_to_include: List[str],
    end_activities_to_include: List[str],
) -> Tuple[List[str], List[Tuple[str, str]]]:
    """
    Sort the edges of the directly-follows graph based on reachability principles
    (start activities are putting at the beginning, end activities at the end)

    Parameters
    ----------------
    dfg
        List of edges of the directly-follows graph (without frequency/performance annotation)
    start_activities_to_include
        Start activities
    end_activities
        End activities

    Returns
    ----------------
    sorted_activities
        Activities sorted by reachability
    sorted_edges
        Edges sorted by reachability
    """
    # identify all unique activities
    activities_dfg = set(x[0] for x in dfg).union(set(x[1] for x in dfg))

    # create adjacency lists and in-degree count
    adjacency_list = defaultdict(list)
    in_degree = defaultdict(int)

    for u, v in dfg:
        adjacency_list[u].append(v)
        in_degree[v] += 1
        if u not in in_degree:
            in_degree[u] = 0

    # initialize the queue with start activities
    queue = deque(start_activities_to_include)
    distance = {activity: 0 for activity in start_activities_to_include}

    # ensure all activities are present in the distance dictionary
    for activity in activities_dfg:
        if activity not in distance:
            distance[activity] = float("inf")

    # perform BFS to calculate the distance of each activity from the start
    # activities
    while queue:
        current = queue.popleft()
        current_distance = distance[current]

        for neighbor in adjacency_list[current]:
            if distance[neighbor] > current_distance + 1:
                distance[neighbor] = current_distance + 1
                queue.append(neighbor)

    # sort edges based on the distance of their source activities
    def edge_priority(edge):
        u, v = edge
        if u in start_activities_to_include:
            return (0, distance[u], distance[v], u, v)
        if v in end_activities_to_include:
            return (2, distance[u], distance[v], u, v)
        return (1, distance[u], distance[v], u, v)

    sorted_edges = sorted(dfg, key=edge_priority)

    # Step 6: Sort activities based on their distance
    sorted_activities = sorted(activities_dfg, key=lambda x: (distance[x], x))

    return sorted_activities, sorted_edges


def graphviz_visualization(
    activities_count,
    dfg,
    image_format="png",
    measure="frequency",
    max_no_of_edges_in_diagram=100000,
    start_activities=None,
    end_activities=None,
    serv_time=None,
    font_size="12",
    bgcolor=constants.DEFAULT_BGCOLOR,
    rankdir=constants.DEFAULT_RANKDIR_GVIZ,
    enable_graph_title: bool = constants.DEFAULT_ENABLE_GRAPH_TITLES,
    graph_title: str = "Directly-Follows Graph",
):
    """
    Do GraphViz visualization of a DFG graph

    Parameters
    -----------
    activities_count
        Count of attributes in the log (may include attributes that are not in the DFG graph)
    dfg
        DFG graph
    image_format
        GraphViz should be represented in this format
    measure
        Describes which measure is assigned to edges in direcly follows graph (frequency/performance)
    max_no_of_edges_in_diagram
        Maximum number of edges in the diagram allowed for visualization
    start_activities
        Start activities of the log
    end_activities
        End activities of the log
    serv_time
        For each activity, the service time in the log
    font_size
        Size of the text on the activities/edges
    bgcolor
        Background color of the visualization (i.e., 'transparent', 'white', ...)
    rankdir
        Direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
    enable_graph_title
        Enables the visualization of a graph's title
    graph_title
        Graph title to display (if enable_graph_title)

    Returns
    -----------
    viz
        Digraph object
    """
    if start_activities is None:
        start_activities = []
    if end_activities is None:
        end_activities = []

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        "",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor, "rankdir": rankdir},
    )

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="'
            + str(2 * int(font_size))
            + '">'
            + graph_title
            + "</FONT>>",
            labelloc="top",
        )

    # first, remove edges in diagram that exceeds the maximum number of edges
    # in the diagram
    dfg_key_value_list = []
    for edge in dfg:
        dfg_key_value_list.append([edge, dfg[edge]])
    # more fine grained sorting to avoid that edges that are below the threshold are
    # undeterministically removed
    dfg_key_value_list = sorted(
        dfg_key_value_list,
        key=lambda x: (x[1], x[0][0], x[0][1]),
        reverse=True,
    )
    dfg_key_value_list = dfg_key_value_list[
        0: min(len(dfg_key_value_list), max_no_of_edges_in_diagram)
    ]
    dfg_allowed_keys = [x[0] for x in dfg_key_value_list]
    dfg_keys = list(dfg.keys())
    for edge in dfg_keys:
        if edge not in dfg_allowed_keys:
            del dfg[edge]

    activities_count_int = copy(activities_count)

    activities_in_dfg = set(activities_count)

    # assign attributes color
    if measure == "frequency":
        activities_color = get_activities_color(activities_count_int)
    else:
        activities_color = get_activities_color_serv_time(serv_time)

    # represent nodes
    viz.attr("node", shape="box")

    if len(activities_in_dfg) == 0:
        activities_to_include = sorted(list(set(activities_count_int)))
    else:
        # take unique elements as a list not as a set (in this way, nodes are
        # added in the same order to the graph)
        activities_to_include = sorted(list(set(activities_in_dfg)))

    start_activities_to_include = [
        act for act in start_activities if act in activities_to_include
    ]
    end_activities_to_include = [
        act for act in end_activities if act in activities_to_include
    ]

    # calculate edges penwidth
    ext_dfg = copy(dfg)
    if start_activities_to_include is not None and start_activities_to_include:
        for sact in start_activities_to_include:
            ext_dfg[(constants.DEFAULT_ARTIFICIAL_START_ACTIVITY, sact)] = (
                start_activities[sact]
            )
    if end_activities_to_include is not None and end_activities_to_include:
        for eact in end_activities_to_include:
            ext_dfg[(eact, constants.DEFAULT_ARTIFICIAL_END_ACTIVITY)] = (
                end_activities[eact]
            )
    dfg_values = dfg.values()
    min_dfg_value = min(dfg_values)
    max_dfg_value = max(dfg_values)

    penwidth = assign_penwidth_edges(ext_dfg)

    dfg_edges = sorted(list(dfg.keys()))
    if start_activities_to_include and end_activities_to_include:
        activities_to_include, dfg_edges = sort_dfg_reachability(
            dfg_edges, start_activities_to_include, end_activities_to_include
        )

    activities_map = {}

    for act in activities_to_include:
        if "frequency" in measure and act in activities_count_int:
            viz.node(
                str(hash(act)),
                act + " (" + str(activities_count_int[act]) + ")",
                style="filled",
                fillcolor=activities_color[act],
                fontsize=font_size,
            )
            activities_map[act] = str(hash(act))
        elif (
            "performance" in measure
            and act in serv_time
            and serv_time[act] >= 0
        ):
            viz.node(
                str(hash(act)),
                act + " (" + human_readable_stat(serv_time[act]) + ")",
                fontsize=font_size,
                style="filled",
                fillcolor=activities_color[act],
            )
            activities_map[act] = str(hash(act))
        else:
            viz.node(str(hash(act)), act, fontsize=font_size)
            activities_map[act] = str(hash(act))

    # represent edges
    for edge in dfg_edges:
        if "frequency" in measure or "cost" in measure:
            label = str(dfg[edge])
        else:
            label = human_readable_stat(dfg[edge])

        color = None
        if "performance" in measure:
            color = value_to_color(dfg[edge], min_dfg_value, max_dfg_value)

        viz.edge(
            str(hash(edge[0])),
            str(hash(edge[1])),
            label=label,
            penwidth=str(penwidth[edge]),
            fontsize=font_size,
            color=color,
        )

    if start_activities_to_include:
        viz.node("@@startnode", constants.DEFAULT_START_SYMBOL_GRAPHS, shape="circle", fontsize="34")
        for act in start_activities_to_include:
            label = (
                str(start_activities[act])
                if isinstance(start_activities, dict)
                and measure == "frequency"
                else ""
            )
            viz.edge(
                "@@startnode",
                activities_map[act],
                label=label,
                fontsize=font_size,
                penwidth=str(
                    penwidth[
                        (constants.DEFAULT_ARTIFICIAL_START_ACTIVITY, act)
                    ]
                ),
            )

    if end_activities_to_include:
        viz.node("@@endnode", constants.DEFAULT_END_SYMBOL_GRAPHS, shape="doublecircle", fontsize="32")
        for act in end_activities_to_include:
            label = (
                str(end_activities[act])
                if isinstance(end_activities, dict) and measure == "frequency"
                else ""
            )
            viz.edge(
                activities_map[act],
                "@@endnode",
                label=label,
                fontsize=font_size,
                penwidth=str(
                    penwidth[(act, constants.DEFAULT_ARTIFICIAL_END_ACTIVITY)]
                ),
            )

    viz.attr(overlap="false")
    viz.attr(fontsize="11")

    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

import pandas as pd

from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventStream
from pm4py.util import constants
from pm4py.util import exec_utils
from pm4py.util import points_subset
from pm4py.util import xes_constants, pandas_utils
from pm4py.util.dt_parsing.variants import strpfromiso
import random
import traceback


LEGACY_PARQUET_TP_REPLACER = "AAA"
LEGACY_PARQUET_CASECONCEPTNAME = "caseAAAconceptAAAname"


class Parameters(Enum):
    PARTITION_COLUMN = "partition_column"
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    CASE_PREFIX = constants.CASE_ATTRIBUTE_PREFIX
    CASE_ATTRIBUTES = "case_attributes"
    CONSIDER_ALL_ATTRIBUTES = "consider_all_attributes"
    MANDATORY_ATTRIBUTES = "mandatory_attributes"
    MAX_NO_CASES = "max_no_cases"
    MIN_DIFFERENT_OCC_STR_ATTR = "min_different_occ_str_attr"
    MAX_DIFFERENT_OCC_STR_ATTR = "max_different_occ_str_attr"
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    PARAM_ARTIFICIAL_START_ACTIVITY = constants.PARAM_ARTIFICIAL_START_ACTIVITY
    PARAM_ARTIFICIAL_END_ACTIVITY = constants.PARAM_ARTIFICIAL_END_ACTIVITY
    INDEX_KEY = "index_key"
    CASE_INDEX_KEY = "case_index_key"
    USE_EXTREMES_TIMESTAMP = "use_extremes_timestamp"
    ADD_CASE_IDENTIFIER_COLUMN = "add_case_identifier_column"
    DETERMINISTIC = "deterministic"
    COUNT_OCCURRENCES = "count_occurrences"


from pm4py.objects.log.util.df_features_utils import *  # noqa: F401,F403


def insert_partitioning(df, num_partitions, parameters=None):
    """
    Insert the partitioning in the specified dataframe

    Parameters
    -------------
    df
        Dataframe
    num_partitions
        Number of partitions
    parameters
        Parameters of the algorithm

    Returns
    -------------
    df
        Partitioned dataframe
    """
    if parameters is None:
        parameters = {}

    case_index_key = exec_utils.get_param_value(
        Parameters.CASE_INDEX_KEY, parameters, constants.DEFAULT_CASE_INDEX_KEY
    )
    partition_column = exec_utils.get_param_value(
        Parameters.PARTITION_COLUMN, parameters, "@@partitioning"
    )

    if case_index_key not in df.columns:
        from pm4py.util import pandas_utils

        df = pandas_utils.insert_case_index(df, case_index_key)

    df[partition_column] = df[case_index_key] % num_partitions

    return df


def legacy_parquet_support(df, parameters=None):
    """
    For legacy support, Parquet files columns could not contain
    a ":" that has been arbitrarily replaced by a replacer string.
    This string substitutes the replacer to the :

    Parameters
    ---------------
    dataframe
        Dataframe
    parameters
        Parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    df.columns = [
        x.replace(LEGACY_PARQUET_TP_REPLACER, ":") for x in df.columns
    ]

    return df


def table_to_stream(table, parameters=None):
    """
    Converts a Pyarrow table to an event stream

    Parameters
    ------------
    table
        Pyarrow table
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    dict0 = table.to_pydict()
    keys = list(dict0.keys())
    # for legacy format support
    if LEGACY_PARQUET_CASECONCEPTNAME in keys:
        for key in keys:
            dict0[key.replace(LEGACY_PARQUET_TP_REPLACER, ":")] = dict0.pop(
                key
            )

    stream = EventStream([dict(zip(dict0, i)) for i in zip(*dict0.values())])

    return stream


def table_to_log(table, parameters=None):
    """
    Converts a Pyarrow table to an event log

    Parameters
    ------------
    table
        Pyarrow table
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    stream = table_to_stream(table, parameters=parameters)

    return log_converter.apply(stream, parameters=parameters)


def convert_timestamp_columns_in_df(
    df, timest_format=None, timest_columns=None
):
    """
    Convert all dataframe columns in a dataframe

    Parameters
    -----------
    df
        Dataframe
    timest_format
        (If provided) Format of the timestamp columns in the CSV file
    timest_columns
        Columns of the CSV that shall be converted into timestamp

    Returns
    ------------
    df
        Dataframe with timestamp columns converted

    """
    if timest_format is None:
        timest_format = constants.DEFAULT_TIMESTAMP_PARSE_FORMAT

    if timest_format is None:
        timest_format = "mixed"

    for col in df.columns:
        if timest_columns is None or col in timest_columns:
            if "obj" in str(df[col].dtype) or "str" in str(df[col].dtype):
                try:
                    df[col] = pandas_utils.dataframe_column_string_to_datetime(
                        df[col], format=timest_format, utc=True
                    )
                except BaseException:
                    try:
                        df[col] = (
                            pandas_utils.dataframe_column_string_to_datetime(
                                df[col],
                                format=timest_format,
                                exact=False,
                                utc=True,
                            )
                        )
                    except BaseException:
                        try:
                            df[col] = (
                                pandas_utils.dataframe_column_string_to_datetime(
                                    df[col], format=timest_format))
                        except BaseException:
                            pass

    for col in df.columns:
        if "date" in str(df[col].dtype) or "time" in str(df[col].dtype):
            df[col] = strpfromiso.fix_dataframe_column(df[col])

    return df


def sample_dataframe(df, parameters=None):
    """
    Sample a dataframe on a given number of cases

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY
        - Parameters.CASE_ID_TO_RETAIN

    Returns
    -------------
    sampled_df
        Sampled dataframe
    """
    if parameters is None:
        parameters = {}

    deterministic = exec_utils.get_param_value(
        Parameters.DETERMINISTIC, parameters, False
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    max_no_cases = exec_utils.get_param_value(
        Parameters.MAX_NO_CASES, parameters, 100
    )

    case_ids = pandas_utils.format_unique(df[case_id_key].unique())
    case_ids = list(case_ids)

    if not deterministic:
        random.shuffle(case_ids)

    case_id_to_retain = points_subset.pick_chosen_points_list(
        min(max_no_cases, len(case_ids)), case_ids
    )

    return df[df[case_id_key].isin(case_id_to_retain)]


def insert_artificial_start_end(
    df0: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> pd.DataFrame:
    """
    Inserts the artificial start/end activities in a Pandas dataframe

    Parameters
    ------------------
    df0
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY: the case identifier
        - Parameters.TIMESTAMP_KEY: the timestamp
        - Parameters.ACTIVITY_KEY: the activity

    Returns
    -----------------
    enriched_df
        Dataframe with artificial start/end activities
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    use_extremes_timestamp = exec_utils.get_param_value(
        Parameters.USE_EXTREMES_TIMESTAMP, parameters, False
    )

    artificial_start_activity = exec_utils.get_param_value(
        Parameters.PARAM_ARTIFICIAL_START_ACTIVITY,
        parameters,
        constants.DEFAULT_ARTIFICIAL_START_ACTIVITY,
    )
    artificial_end_activity = exec_utils.get_param_value(
        Parameters.PARAM_ARTIFICIAL_END_ACTIVITY,
        parameters,
        constants.DEFAULT_ARTIFICIAL_END_ACTIVITY,
    )

    index_key = exec_utils.get_param_value(
        Parameters.INDEX_KEY, parameters, constants.DEFAULT_INDEX_KEY
    )

    df = df0.copy()
    df = pandas_utils.insert_index(df, index_key)
    df = df.sort_values([case_id_key, timestamp_key, index_key])

    start_df = (
        df[[case_id_key, timestamp_key]]
        .groupby(case_id_key)
        .first()
        .reset_index()
    )
    end_df = (
        df[[case_id_key, timestamp_key]]
        .groupby(case_id_key)
        .last()
        .reset_index()
    )
    # stability trick: remove 1ms from the artificial start activity
    # timestamp, add 1ms to the artificial end activity timestamp
    if use_extremes_timestamp:
        start_df[timestamp_key] = pd.Timestamp.min
        end_df[timestamp_key] = pd.Timestamp.max
        start_df[timestamp_key] = start_df[timestamp_key].dt.tz_localize("utc")
        end_df[timestamp_key] = end_df[timestamp_key].dt.tz_localize("utc")
    else:
        start_df[timestamp_key] = start_df[timestamp_key] - pd.Timedelta(
            "1 ms"
        )
        end_df[timestamp_key] = end_df[timestamp_key] + pd.Timedelta("1 ms")

    start_df[activity_key] = artificial_start_activity
    end_df[activity_key] = artificial_end_activity

    df = pandas_utils.concat([start_df, df, end_df])
    df = pandas_utils.insert_index(df, index_key)
    df = df.sort_values([case_id_key, timestamp_key, index_key])

    df.attrs = df0.attrs

    return df


def dataframe_to_activity_case_table(
    df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
):
    """
    Transforms a Pandas dataframe into:
    - an "activity" table, containing the events and their attributes
    - a "case" table, containing the cases and their attributes

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm that should be used, including:
        - Parameters.CASE_ID_KEY => the column to be used as case ID (shall be included both in the activity table and the case table)
        - Parameters.CASE_PREFIX => if a list of attributes at the case level is not provided, then all the ones of the dataframe
                                    starting with one of these are considered.
        - Parameters.CASE_ATTRIBUTES => the attributes of the dataframe to be used as case columns

    Returns
    ---------------
    activity_table
        Activity table
    case_table
        Case table
    """
    if parameters is None:
        parameters = {}

    # make sure we start from a dataframe object
    df = log_converter.apply(
        df, variant=log_converter.Variants.TO_DATA_FRAME, parameters=parameters
    )

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    case_id_prefix = exec_utils.get_param_value(
        Parameters.CASE_PREFIX, parameters, constants.CASE_ATTRIBUTE_PREFIX
    )

    case_attributes = exec_utils.get_param_value(
        Parameters.CASE_ATTRIBUTES,
        parameters,
        set([x for x in df.columns if x.startswith(case_id_prefix)]),
    )
    event_attributes = set([x for x in df.columns if x not in case_attributes])

    activity_table = df[event_attributes.union({case_id_key})]
    case_table = (
        df[case_attributes.union({case_id_key})]
        .groupby(case_id_key)
        .first()
        .reset_index()
    )

    return activity_table, case_table



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
from copy import deepcopy, copy
from enum import Enum
from typing import Optional, Dict, Any, Union, Tuple

import numpy as np
import pandas as pd

from pm4py.algo.conformance.alignments.petri_net import algorithm as ali
from pm4py.algo.conformance.alignments.petri_net.variants import (
    state_equation_a_star as star,
)
from pm4py.algo.conformance.tokenreplay import algorithm as token_replay
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net import properties as petri_properties
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.statistics.attributes.log.select import (
    select_attributes_from_log_for_tree,
)
from pm4py.statistics.variants.log import get as variants_module
from pm4py.util import constants, xes_constants
from pm4py.util import exec_utils, pandas_utils
from pm4py.visualization.decisiontree.util import dt_to_string


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    LABELS = "labels"


def create_data_petri_nets_with_decisions(
        log: Union[EventLog, pd.DataFrame],
        net: PetriNet,
        initial_marking: Marking,
        final_marking: Marking,
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Given a Petri net, create a data Petri net with the decisions
    given for each place by the decision mining algorithm.

    Parameters
    ----------------
    log
        Event log (EventLog or DataFrame).
    net
        Petri net.
    initial_marking
        Initial marking of the Petri net.
    final_marking
        Final marking of the Petri net.

    Returns
    ------------------
    data_petri_net
        Petri net enriched with guards (conditions).
    initial_marking
        Initial marking (unchanged).
    final_marking
        Final marking (unchanged).
    """
    all_conditions = {}
    all_variables = {}

    # Attempt to build a decision tree for each place in the net
    for place in net.places:
        try:
            dt_classifier, columns, targets = get_decision_tree(
                log,
                net,
                initial_marking,
                final_marking,
                decision_point=place.name,
                parameters={"labels": False},  # by default
            )
            # Convert the scikit-learn tree into string-based conditions
            target_classes_dict, variables_dict = dt_to_string.apply(
                dt_classifier, columns
            )
            # Remap integer-based classes to the actual transitions
            target_classes = {
                targets[int(k)]: v for k, v in target_classes_dict.items()
            }
            variables = {targets[int(k)]: v for k, v in variables_dict.items()}

            # Accumulate results
            for transition_name in target_classes.keys():
                all_conditions[transition_name] = target_classes[transition_name]
                all_variables[transition_name] = variables[transition_name]

        except Exception:
            # If the decision-tree building fails for some place, ignore it
            pass

    # Attach discovered guards (conditions) to transitions
    for transition in net.transitions:
        if transition.name in all_conditions:
            transition.properties[petri_properties.TRANS_GUARD] = all_conditions[
                transition.name
            ]
            transition.properties[petri_properties.READ_VARIABLE] = all_variables[
                transition.name
            ]
            transition.properties[petri_properties.WRITE_VARIABLE] = []

    return net, initial_marking, final_marking


def get_decision_tree(
        log: Union[EventLog, pd.DataFrame],
        net: PetriNet,
        initial_marking: Marking,
        final_marking: Marking,
        decision_point=None,
        attributes=None,
        parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Any:
    """
    Gets a decision tree classifier on a specific point of the model.

    Parameters
    --------------
    log
        Event log (EventLog or DataFrame).
    net
        Petri net.
    initial_marking
        Initial marking.
    final_marking
        Final marking.
    decision_point
        Name of the place in which a decision happens:
        - if not specified, the method raises an Exception with a list of possible decision points.
    attributes
        Attributes of the log. If not specified, an automatic attribute selection is performed.
    parameters
        Parameters of the algorithm.

    Returns
    ---------------
    clf
        Fitted decision tree classifier.
    feature_names
        The names of the features used to fit the classifier.
    classes
        The classes (i.e., transitions) the classifier distinguishes.
    """
    from pm4py.util import ml_utils

    if parameters is None:
        parameters = {}

    # Compute feature matrix X, labels y, and the mapping to actual transitions
    X, y, targets = apply(
        log,
        net,
        initial_marking,
        final_marking,
        decision_point=decision_point,
        attributes=attributes,
        parameters=parameters,
    )

    # Fit a decision tree classifier
    dt_classifier = ml_utils.DecisionTreeClassifier()
    dt_classifier = dt_classifier.fit(X, y)

    # Return the classifier, the feature names, and the classes
    return dt_classifier, list(X.columns.values.tolist()), targets


def apply(
        log: Union[EventLog, pd.DataFrame],
        net: PetriNet,
        initial_marking: Marking,
        final_marking: Marking,
        decision_point=None,
        attributes=None,
        parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Any:
    """
    Gets the essential information (features, target class, and names of the target class)
    in order to learn a classifier.

    Parameters
    --------------
    log
        Event log (EventLog or DataFrame).
    net
        Petri net.
    initial_marking
        Initial marking.
    final_marking
        Final marking.
    decision_point
        The name of the place in which a decision happens.
        - If not specified, raises an Exception with a list of possible places.
    attributes
        Attributes of the log. If not specified, an automatic attribute selection is performed.
    parameters
        Parameters of the algorithm.

    Returns
    ---------------
    X
        DataFrame of features.
    y
        Series of encoded target classes (integer).
    class_name
        Mapping of integer class -> actual transition name.
    """
    if parameters is None:
        parameters = {}

    # Retrieve parameter about labeling
    labels = exec_utils.get_param_value(Parameters.LABELS, parameters, True)

    # Define the default activity key
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    # If decision_point is None, we provide a list of valid places and raise an Exception
    if decision_point is None:
        decision_points_names = get_decision_points(net, labels=labels, parameters=parameters)
        raise Exception(
            "Please provide 'decision_point'. Possible decision points: %s" % list(decision_points_names.keys())
        )

    # If attributes not specified, automatically select them
    if attributes is None:
        str_tr_attr, str_ev_attr, num_tr_attr, num_ev_attr = select_attributes_from_log_for_tree(log)
        attributes = list(str_ev_attr) + list(num_ev_attr)

    # Build the main "decisions table" from the log and the Petri net
    decision_info, _ = get_decisions_table(
        log,
        net,
        initial_marking,
        final_marking,
        attributes=attributes,
        pre_decision_points=[decision_point],
        parameters=parameters,
    )

    # For the chosen decision point, separate attributes (X) and target transitions (y)
    relevant_data = decision_info[decision_point]
    x_attributes = [a for a in attributes if a != activity_key]

    # We separate string and numeric attributes to handle get_dummies for strings
    str_attributes = set()
    non_str_attributes = set()

    # Lists to store intermediate results
    all_feature_dicts_str = []
    all_feature_dicts_num = []
    all_targets = []

    # For each recorded decision, collect attributes
    for (attr_dict, chosen_transition) in relevant_data:
        # Identify which attributes are string vs numeric
        for attr_name, attr_value in attr_dict.items():
            if attr_name in x_attributes:
                if isinstance(attr_value, str):
                    str_attributes.add(attr_name)
                else:
                    non_str_attributes.add(attr_name)

        # Build two partial dicts:
        # one for string columns, one for numeric columns
        all_feature_dicts_str.append({
            a: v for a, v in attr_dict.items()
            if a in x_attributes and isinstance(v, str)
        })
        all_feature_dicts_num.append({
            a: v for a, v in attr_dict.items()
            if a in x_attributes and not isinstance(v, str)
        })

        # The target is the chosen transition's label/name
        all_targets.append(chosen_transition)

    # Convert to pandas DataFrame, applying get_dummies to string features
    df_str = pandas_utils.instantiate_dataframe(all_feature_dicts_str)
    df_str = pd.get_dummies(data=df_str, columns=list(str_attributes))

    df_num = pandas_utils.instantiate_dataframe(all_feature_dicts_num)
    X = pd.concat([df_str, df_num], axis=1)

    Y = pd.DataFrame(all_targets, columns=["Name"])
    Y, targets = encode_target(Y, "Name")  # encode the target transitions
    y = Y["Target"]

    return X, y, targets


def get_decisions_table(
        log0,
        net,
        initial_marking,
        final_marking,
        attributes=None,
        use_trace_attributes=False,
        k=1,
        pre_decision_points=None,
        trace_attributes=None,
        parameters=None,
):
    """
    Builds a decision table out of a log and an accepting Petri net.

    For each place that has multiple outgoing arcs (a "decision point"),
    we record the attributes that preceded the choice of a particular transition.

    Parameters
    -----------------
    log0
        Event log (EventLog or DataFrame).
    net
        Petri net.
    initial_marking
        Initial marking.
    final_marking
        Final marking.
    attributes
        List of event attributes to consider (if not provided, all are considered).
    use_trace_attributes
        Whether to include trace attributes (e.g., case-level data) in the decision table.
    k
        Number of last events to look back at for each decision. (Default=1)
    pre_decision_points
        List of place names that should be considered. If None, the code infers them automatically.
    trace_attributes
        List of trace attribute names to consider. If None, all are considered (if use_trace_attributes=True).
    parameters
        Additional parameters (e.g., {Parameters.LABELS: True/False}).

    Returns
    --------------
    I
        A dictionary keyed by place name. Values are lists of tuples (dict_of_attributes, chosen_transition).
    decision_points
        The dictionary of decision points (places with multiple outgoing arcs),
        possibly filtered by `pre_decision_points`.
    """
    if parameters is None:
        parameters = {}

    # Convert input log to pm4py EventLog if it's a DataFrame
    log = log_converter.apply(log0, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters)

    # Validate some of the inputs
    _validate_inputs_for_decision_table(
        pre_decision_points, attributes, use_trace_attributes, trace_attributes
    )

    # Identify actual decision points
    # (places that have >=2 outgoing arcs)
    decision_points = get_decision_points(
        net, pre_decision_points=pre_decision_points, parameters=parameters
    )
    # Also get decision point labels if needed
    decision_points_names = get_decision_points(
        net, labels=True, pre_decision_points=pre_decision_points, parameters=parameters
    )

    # Possibly rename trace/event attributes to avoid collisions
    if use_trace_attributes:
        log = prepare_event_log(log)
        if attributes is not None:
            attributes = prepare_attributes(attributes)

    # If no explicit trace_attributes are given, collect from the log
    if use_trace_attributes and trace_attributes is None:
        trace_attributes = set()
        for trace in log:
            trace_attributes.update(trace.attributes.keys())
        trace_attributes = list(trace_attributes)

    # If attributes are still None, collect all event attributes from the log
    if attributes is None:
        attributes = set()
        for trace in log:
            for event in trace:
                attributes.update(list(event.keys()))
        attributes = list(attributes)

    # Now build the dictionary of decision info
    # Each key is a place name, and the value is a list of (attributes_dict, chosen_transition)
    decision_info = _build_decision_info(
        log,
        net,
        initial_marking,
        final_marking,
        decision_points,
        decision_points_names,
        attributes,
        use_trace_attributes,
        trace_attributes,
        k,
        parameters,
    )
    return (decision_info, decision_points)


def prepare_event_log(log):
    """
    If trace attributes are considered, we want to differentiate them from event attributes.
    For trace attributes, we prepend "t_".
    For event attributes, we prepend "e_".

    This helps avoid collisions when both trace and event attributes share the same name.

    Parameters
    ----------
    log : EventLog
        The original log.

    Returns
    -------
    EventLog
        The modified log with attribute names prefixed.
    """
    for trace in log:
        # Prefix trace attributes with "t_"
        trace_attrs_copy = dict(trace.attributes)
        for attribute in trace_attrs_copy:
            new_key = "t_" + attribute
            trace.attributes[new_key] = trace.attributes.pop(attribute)

        # Prefix event attributes with "e_"
        for event in trace:
            event_attrs_copy = dict(event._dict)
            for attribute in event_attrs_copy:
                new_key = "e_" + attribute
                event._dict[new_key] = event._dict.pop(attribute)

    return log


def prepare_attributes(attributes):
    """
    If trace attributes are considered, we assume all the user-provided attributes refer to event attributes
    and prepend "e_" to them.

    Parameters
    ----------
    attributes : list
        List of original attribute names.

    Returns
    -------
    list
        List of attribute names, each prefixed by "e_".
    """
    new_attributes = []
    for attribute in attributes:
        new_attributes.append("e_" + attribute)
    return new_attributes


def get_decision_points(
        net, labels=False, pre_decision_points=None, parameters=None
):
    """
    Identifies "decision points" in the net, i.e., places with >= 2 outgoing arcs.

    Parameters
    ----------
    net : PetriNet
        The Petri net under analysis.
    labels : bool
        Whether to list the labels of transitions as values rather than the raw transition names.
    pre_decision_points : list or None
        If provided, only return decision points that appear in this list (filter).
    parameters : dict
        (Unused in this function except for consistency.)

    Returns
    -------
    dict
        A dictionary mapping place_name -> list of outgoing transition names or labels.
    """
    if parameters is None:
        parameters = {}

    # Build a dict {place_name -> [outgoing transitions]}
    outgoing_dict = {}
    for place in net.places:
        outgoing_dict[place.name] = []

    # For each arc that starts from a place, record the target transition
    for arc in net.arcs:
        if arc.source in net.places:
            transition_label = arc.target.label if labels else arc.target.name
            outgoing_dict[arc.source.name].append(transition_label)

    # Filter only those with >=2 outgoing arcs
    decision_points = {
        place_name: trans_list
        for place_name, trans_list in outgoing_dict.items()
        if len(trans_list) >= 2
    }

    # If user gave a specific list of decision points, filter further
    if pre_decision_points is not None:
        _check_pre_decision_points(decision_points, pre_decision_points)

    return decision_points


def get_attributes(
        log,
        decision_points,
        attributes,
        use_trace_attributes,
        trace_attributes,
        k,
        net,
        initial_marking,
        final_marking,
        decision_points_names,
        parameters=None,
):
    """
    For each decision place, this collects the final table of
    (attributes -> chosen transition) for each occurrence of a decision.

    This function internally uses token-based replay (or alignment for non-fitting traces)
    to discover the actual transitions that were used from the log.
    Then, for each place with multiple outgoing arcs, we store the attributes
    that led to a certain chosen transition.

    Parameters
    ----------
    log : EventLog
        The event log.
    decision_points : dict
        Dictionary mapping place_name -> list of possible transitions (IDs/names).
    attributes : list
        Attributes to consider from events.
    use_trace_attributes : bool
        Whether to consider trace-level attributes as well.
    trace_attributes : list
        List of trace-level attributes to consider.
    k : int
        Number of events to look back at each decision (the "window size").
    net : PetriNet
        The Petri net.
    initial_marking : Marking
        Initial marking.
    final_marking : Marking
        Final marking.
    decision_points_names : dict
        Dictionary mapping place_name -> list of transition labels (if labels=True).
    parameters : dict
        Additional parameters (e.g. {Parameters.LABELS: True/False}).

    Returns
    -------
    dict
        A dictionary keyed by place name, with each value a list of tuples:
        (attributes_dict, chosen_transition).
    """
    if parameters is None:
        parameters = {}

    # Check if transitions should use .label or .name
    labels = exec_utils.get_param_value(Parameters.LABELS, parameters, True)

    # Prepare the result structure
    decision_info = {place_name: [] for place_name in decision_points.keys()}

    # For collecting variants/traces
    variants_idxs = variants_module.get_variants_from_log_trace_idx(log, parameters=parameters)
    unique_variants = list(variants_idxs.keys())

    filtered_log = EventLog()
    for variant_name in variants_idxs:
        filtered_log.append(log[variants_idxs[variant_name][0]])

    # Token replay over the entire log
    replay_result = token_replay.apply(
        filtered_log, net, initial_marking, final_marking, parameters=parameters
    )

    # For each replayed variant
    for variant_idx, variant_name in enumerate(unique_variants):
        # Check if the replayed variant is perfectly fitting or not
        token_replay_entry = replay_result[variant_idx]

        # For each trace in this variant
        trace_indices = variants_idxs[variant_name]

        if token_replay_entry["trace_fitness"] == 1.0:
            # Perfectly fitting; we can simply rely on "activated_transitions"
            _extract_decisions_perfect_fit(
                log, trace_indices, token_replay_entry, decision_info,
                decision_points, decision_points_names, attributes,
                use_trace_attributes, trace_attributes, k, labels
            )
        else:
            # Not a perfect fit; we must do alignments for each trace in this variant
            # (We align only for the first example trace in that variant to get the alignment.)
            sample_trace_idx = trace_indices[0]
            sample_trace = log[sample_trace_idx]

            alignment_params = copy(parameters)
            alignment_params[star.Parameters.PARAM_ALIGNMENT_RESULT_IS_SYNC_PROD_AWARE] = True
            alignment_info = ali.apply(
                sample_trace, net, initial_marking, final_marking, parameters=alignment_params
            )

            # Now apply the same alignment to each trace in the variant
            for idx in trace_indices:
                _extract_decisions_alignment(
                    log, idx, alignment_info, decision_info, decision_points,
                    attributes, use_trace_attributes, trace_attributes, k, labels
                )

    return decision_info


def encode_target(df, target_column):
    """
    Adds a 'Target' column to df with integer-encoded classes
    derived from an existing column (target_column).

    Method adapted from:
    http://chrisstrelioff.ws/sandbox/2015/06/08/decision_trees_in_python_with_scikit_learn_and_pandas.html

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame containing the target_column.
    target_column : str
        The name of the column to map to integer classes.

    Returns
    -------
    (df_mod, targets)
        df_mod is the modified DataFrame with a 'Target' column.
        targets is the list of unique target names in their mapped order.
    """
    df_mod = df.copy()
    unique_targets = pandas_utils.format_unique(df_mod[target_column].unique())
    map_to_int = {name: n for n, name in enumerate(unique_targets)}
    df_mod["Target"] = df_mod[target_column].replace(map_to_int)
    return df_mod, unique_targets


# ----------------------------------------------------------------
#                          HELPERS
# ----------------------------------------------------------------

def _validate_inputs_for_decision_table(
        pre_decision_points, attributes, use_trace_attributes, trace_attributes
):
    """
    Internal function to validate the input parameters for get_decisions_table,
    giving more user-friendly error messages.
    """
    # Validate pre_decision_points
    if pre_decision_points is not None:
        if not isinstance(pre_decision_points, list):
            raise ValueError(
                "pre_decision_points must be a list of place names (strings)."
            )
        if len(pre_decision_points) == 0:
            raise ValueError(
                "The list pre_decision_points cannot be empty if provided."
            )

    # Validate attributes
    if attributes is not None:
        if not isinstance(attributes, list):
            raise ValueError(
                "attributes must be a list of event attribute names (strings)."
            )
        if len(attributes) == 0:
            raise ValueError(
                "The list attributes cannot be empty if provided."
            )

    # If trace attributes are provided, force use_trace_attributes=True
    if trace_attributes is not None:
        if not isinstance(trace_attributes, list):
            raise ValueError(
                "trace_attributes must be a list of trace attribute names (strings)."
            )
        if len(trace_attributes) == 0:
            raise ValueError(
                "The list trace_attributes cannot be empty if provided."
            )
        if not use_trace_attributes:
            # Warn user, then override
            print(
                "Note: 'trace_attributes' list given, but use_trace_attributes=False. "
                "Enabling use_trace_attributes=True."
            )


def _check_pre_decision_points(decision_points, pre_decision_points):
    """
    Checks which pre_decision_points are actually valid (places with >=2 outgoing arcs).
    Removes from 'decision_points' any place not in 'pre_decision_points',
    and warns if some user-specified place is not a decision point.
    """
    valid_count = 0
    to_remove = []
    for dp in list(decision_points.keys()):
        if dp not in pre_decision_points:
            to_remove.append(dp)
        else:
            valid_count += 1

    # Remove not requested places
    for dp in to_remove:
        del decision_points[dp]

    # Warn if none are valid
    if valid_count == 0:
        raise Exception("None of the given places is a decision point.")

    # Warn if some were invalid
    if valid_count < len(pre_decision_points):
        print(
            "Warning: Some of the places in pre_decision_points are not actual decision points "
            "in the Petri net. Only valid ones have been retained."
        )


def _build_decision_info(
        log,
        net,
        initial_marking,
        final_marking,
        decision_points,
        decision_points_names,
        attributes,
        use_trace_attributes,
        trace_attributes,
        k,
        parameters
):
    """
    Internal function to orchestrate the extraction of attributes that lead to particular transition choices.
    Returns a dictionary keyed by place name with a list of (attributes_dict, chosen_transition).
    """
    # Token replay or align each trace to find the sequences of transitions.
    # Then, for each place that has multiple outgoing arcs, store the attributes
    # that were present before the transition was fired.

    # We re-use get_attributes() because it contains the logic for
    # collecting attributes from the replay or alignments.
    return get_attributes(
        log=log,
        decision_points=decision_points,
        attributes=attributes,
        use_trace_attributes=use_trace_attributes,
        trace_attributes=trace_attributes,
        k=k,
        net=net,
        initial_marking=initial_marking,
        final_marking=final_marking,
        decision_points_names=decision_points_names,
        parameters=parameters,
    )


def _extract_decisions_perfect_fit(
    log,
    trace_indices,
    token_replay_entry,
    decision_info,
    decision_points,
    decision_points_names,
    attributes,
    use_trace_attributes,
    trace_attributes,
    k,
    labels
):
    """
    Extract decisions from a token replay entry with trace_fitness == 1.0 (perfect fit),
    ensuring that we only use attributes known BEFORE firing each transition.
    """
    for trace_index in trace_indices:
        # We'll keep a rolling window of the last k attributes
        window_attributes = [None] * k

        trace = log[trace_index]
        # If we use trace attributes, set them once for the entire trace
        global_attrs = {}
        if use_trace_attributes and trace_attributes:
            for attribute in trace_attributes:
                if attribute in trace.attributes:
                    global_attrs[attribute] = trace.attributes[attribute]

        # Pointer to the current event in the trace
        event_idx = 0

        # For each transition in the replay
        for transition in token_replay_entry["activated_transitions"]:
            # (1) Store decision with the current window (i.e., no new event has been consumed yet)
            transition_name_or_label = transition.label if labels else transition.name
            for place_name, outgoings in decision_points_names.items():
                if transition_name_or_label in outgoings:
                    # We have a decision at place_name
                    for attr_dict in window_attributes:
                        if attr_dict is not None:
                            # Record the known attributes for the chosen transition
                            decision_info[place_name].append(
                                (attr_dict.copy(), transition_name_or_label)
                            )

            # (2) If this transition corresponds to a visible event (transition.label != None),
            #     then consume the *next* event from the log to update the window
            #     so that it is NOT used to decide the *current* transition.
            if transition.label is not None and event_idx < len(trace):
                # Merge global (trace) attributes + event attributes for the next event
                attr_dict = dict(global_attrs)
                for attr_name in attributes:
                    if attr_name in trace[event_idx]:
                        attr_dict[attr_name] = trace[event_idx][attr_name]

                # Put it in the rolling window
                window_attributes[event_idx % k] = attr_dict.copy()

                # Advance the pointer to the next event
                event_idx += 1


def _extract_decisions_alignment(
        log,
        trace_index,
        alignment_info,
        decision_info,
        decision_points,
        attributes,
        use_trace_attributes,
        trace_attributes,
        k,
        labels
):
    """
    Extract decisions for a single trace using alignment information.
    """
    # The alignment is a list of moves, each move is a tuple:
    # ((model_name, model_label), (log_name, log_label))
    alignment = alignment_info["alignment"]
    trace = log[trace_index]

    # Collect trace-level attributes if requested
    global_attrs = {}
    if use_trace_attributes and trace_attributes:
        for attribute in trace_attributes:
            if attribute in trace.attributes:
                global_attrs[attribute] = trace.attributes[attribute]

    # Prepare a rolling window for the last k attributes
    window_attributes = [None] * k

    # Pointer to the current event in this trace
    event_idx = 0

    for trans_names, trans_labels in alignment:
        # Unpack them for clarity
        log_name, model_name = trans_names
        log_label, model_label = trans_labels

        # CASE A: If model_name != '>>', we have a move in the *model* (either synchronous or model-only)
        if model_name != ">>":
            if labels:
                chosen_transition = model_label if model_label is not None else model_name
            else:
                chosen_transition = model_name

            # Now check if that chosen_transition is in some place's outgoing transitions
            # If your decision_points dict is keyed by place_name -> [labels],
            # then you see if chosen_transition is one of those labels
            for place_name, outgoings in decision_points.items():
                if model_name in outgoings:
                    # For each "window" item, store the decision
                    for attr_dict in window_attributes:
                        if attr_dict is not None:
                            # We append a tuple (attribute_dict, chosen_transition)
                            decision_info[place_name].append((attr_dict.copy(), chosen_transition))

        # CASE B: If log_name != '>>', we consumed an event from the log
        # => update the rolling window with that event's attributes
        if log_name != ">>" and event_idx < len(trace):
            combined_attrs = dict(global_attrs)
            event_attrs = trace[event_idx]

            # Copy only the specified attributes
            for attr_name in attributes:
                if attr_name in event_attrs:
                    combined_attrs[attr_name] = event_attrs[attr_name]

            # Insert into the rolling window
            window_attributes[event_idx % k] = combined_attrs
            event_idx += 1



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.util.constants import GROUPED_DATAFRAME
from pm4py.util import exec_utils
from pm4py.util import constants
from enum import Enum
from typing import Optional, Dict, Any, Union
from collections import Counter
import pandas as pd


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    MAX_NO_POINTS_SAMPLE = "max_no_of_points_to_sample"
    KEEP_ONCE_PER_CASE = "keep_once_per_case"


def get_end_activities(
    df: pd.DataFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, int]:
    """
    Get end activities count

    Parameters
    -----------
    df
        Pandas dataframe
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Case ID column in the dataframe
            Parameters.ACTIVITY_KEY -> Column that represents the activity

    Returns
    -----------
    endact_dict
        Dictionary of end activities along with their count
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    grouped_df = (
        parameters[GROUPED_DATAFRAME]
        if GROUPED_DATAFRAME in parameters
        else None
    )

    if grouped_df is None:
        grouped_df = df.groupby(case_id_glue, sort=False)

    endact_dict = dict(
        Counter(grouped_df[activity_key].last().to_numpy().tolist())
    )
    return endact_dict



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
def get_sorted_start_activities_list(start_activities):
    """
    Gets sorted start attributes list

    Parameters
    ----------
    start_activities
        Dictionary of start attributes associated with their count

    Returns
    ----------
    listact
        Sorted start attributes list
    """
    listact = []
    for sa in start_activities:
        listact.append([sa, start_activities[sa]])
    listact = sorted(listact, key=lambda x: x[1], reverse=True)
    return listact


def get_start_activities_threshold(salist, decreasing_factor):
    """
    Get start attributes cutting threshold

    Parameters
    ----------
    salist
        Sorted start attributes list
    decreasing_factor
        Decreasing factor of the algorithm

    Returns
    ---------
    threshold
        Start attributes cutting threshold
    """
    threshold = salist[0][1]
    for i in range(1, len(salist)):
        value = salist[i][1]
        if value > threshold * decreasing_factor:
            threshold = value
    return threshold



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
from typing import Optional, Dict, Any, Union
from sqlite3 import Connection as SQ3_Connection


def apply(
    db: Union[pd.DataFrame, SQ3_Connection],
    parameters: Optional[Dict[Any, Any]] = None,
) -> str:
    """
    Provides a string containing the required process mining domain knowledge for traditional process mining structures
    (in order for the LLM to produce meaningful queries).

    Parameters
    ---------------
    db
        Database
    parameters
        Optional parameters of the method

    Returns
    --------------
    pm_knowledge
        String containing the required process mining knowledge
    """
    if parameters is None:
        parameters = {}

    descr = """
If you need it, the process variant for a case can be obtained as concatenation of the activities of a case.
If you need it, the duration of a case can be obtained as difference between the timestamp of the first and the last event.
    """

    return descr



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from abc import ABC, abstractmethod
from typing import (
    Optional,
    List,
    Collection,
    Any,
    Tuple,
    Generic,
    TypeVar,
    Dict,
)

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure
from pm4py.objects.process_tree.obj import ProcessTree

T = TypeVar("T", bound=IMDataStructure)


class Cut(ABC, Generic[T]):

    @classmethod
    @abstractmethod
    def operator(
        cls, parameters: Optional[Dict[str, Any]] = None
    ) -> ProcessTree:
        pass

    @classmethod
    @abstractmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Collection[Any]]]:
        pass

    @classmethod
    def apply(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[ProcessTree, List[T]]]:
        g = cls.holds(obj, parameters)
        return (
            (cls.operator(), cls.project(obj, g, parameters))
            if g is not None
            else g
        )

    @classmethod
    @abstractmethod
    def project(
        cls,
        obj: T,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[T]:
        """
        Projection of the given data object (Generic type T).
        Returns a corresponding process tree and the projected sub logs according to the identified groups.
        A precondition of the project function is that it holds on the object for the given Object
        """
        pass



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
from typing import Optional, Dict, Any
from pm4py.util import constants, exec_utils
from enum import Enum
from collections.abc import Sequence


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    CASE_ATTRIBUTE_PREFIX = constants.CASE_ATTRIBUTE_PREFIX


class PandasTraceWrapper(Sequence):
    def __init__(
        self,
        dataframe: pd.DataFrame,
        parameters: Optional[Dict[Any, Any]] = None,
    ):
        if parameters is None:
            parameters = {}

        self.parameters = parameters
        self.dataframe = dataframe
        self.case_attribute_prefix = exec_utils.get_param_value(
            Parameters.CASE_ATTRIBUTE_PREFIX,
            parameters,
            constants.CASE_ATTRIBUTE_PREFIX,
        )

        self.attributes = self.dataframe.loc[0].to_dict()
        self.attributes = {
            x.split(self.case_attribute_prefix)[-1]: y
            for x, y in self.attributes.items()
            if x.startswith(self.case_attribute_prefix)
        }

    def __getitem__(self, key):
        if type(key) is slice:
            start = key.start % len(self.dataframe)
            stop = key.stop % len(self.dataframe)
            sli = slice(start, stop - 1, key.step)
            return self.dataframe.loc[sli].to_dict("records")
        key = key % len(self.dataframe)
        return self.dataframe.loc[key].to_dict()

    def __iter__(self):
        return iter(self.dataframe.to_dict("records"))

    def __len__(self):
        return len(self.dataframe)

    def _get_list(self):
        return self.dataframe.to_dict("records")

    _list = property(_get_list)


class PandasLogWrapper(Sequence):
    # permits to iterate over a Pandas dataframe and access its Traces object
    # *without* a conversion to EventLog
    def __init__(
        self,
        dataframe: pd.DataFrame,
        parameters: Optional[Dict[Any, Any]] = None,
    ):
        if parameters is None:
            parameters = {}

        self.parameters = parameters
        self.case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
        )
        self.dataframe = dataframe

        self.grouped_dataframe = self.dataframe.groupby(
            self.case_id_key
        ).groups
        self.keys = list(self.grouped_dataframe)

        self.attributes = {}
        self.extensions = {}
        self.omni_present = {}
        self.classifiers = {}
        self.properties = {}

    def __getitem__(self, key):
        if type(key) is slice:
            start = key.start % len(self.dataframe)
            stop = key.stop % len(self.dataframe)
            sli = slice(start, stop, key.step)
            ret = []
            for x in self.keys[sli]:
                ret.append(
                    PandasTraceWrapper(
                        self.dataframe.loc[self.grouped_dataframe[x]]
                        .copy()
                        .reset_index(),
                        parameters=self.parameters,
                    )
                )
            return ret
        key = key % len(self.grouped_dataframe)
        return PandasTraceWrapper(
            self.dataframe.loc[self.grouped_dataframe[self.keys[key]]]
            .copy()
            .reset_index(),
            parameters=self.parameters,
        )

    def __iter__(self):
        for key in self.keys:
            yield PandasTraceWrapper(
                self.dataframe.loc[self.grouped_dataframe[key]]
                .copy()
                .reset_index(),
                parameters=self.parameters,
            )

    def __len__(self):
        return len(self.grouped_dataframe)

    def _get_list(self):
        ret = []
        for key in self.keys:
            ret.append(
                PandasTraceWrapper(
                    self.dataframe.loc[self.grouped_dataframe[key]]
                    .copy()
                    .reset_index(),
                    parameters=self.parameters,
                )
            )
        return ret

    _list = property(_get_list)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import polars as pl
from enum import Enum

from pm4py.util import exec_utils, constants, xes_constants
from pm4py.util.business_hours import soj_time_business_hours_diff
from typing import Optional, Dict, Any, Union


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    AGGREGATION_MEASURE = "aggregationMeasure"
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def apply(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, float]:
    """
    Gets the service time per activity on a Polars LazyFrame

    Parameters
    --------------
    lf
        Polars LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.START_TIMESTAMP_KEY => start timestamp key
        - Parameters.TIMESTAMP_KEY => timestamp key
        - Parameters.BUSINESS_HOURS => calculates the difference of time based on the business hours, not the total time.
                                        Default: False
        - Parameters.BUSINESS_HOURS_SLOTS =>
        work schedule of the company, provided as a list of tuples where each tuple represents one time slot of business
        hours. One slot i.e. one tuple consists of one start and one end time given in seconds since week start, e.g.
        [
            (7 * 60 * 60, 17 * 60 * 60),
            ((24 + 7) * 60 * 60, (24 + 12) * 60 * 60),
            ((24 + 13) * 60 * 60, (24 + 17) * 60 * 60),
        ]
        meaning that business hours are Mondays 07:00 - 17:00 and Tuesdays 07:00 - 12:00 and 13:00 - 17:00
        - Parameters.AGGREGATION_MEASURE => performance aggregation measure (sum, min, max, mean, median)

    Returns
    --------------
    soj_time_dict
        Service time dictionary
    """
    if parameters is None:
        parameters = {}

    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )
    workcalendar = exec_utils.get_param_value(
        Parameters.WORKCALENDAR,
        parameters,
        constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR,
    )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    aggregation_measure = exec_utils.get_param_value(
        Parameters.AGGREGATION_MEASURE, parameters, "mean"
    )

    if business_hours:
        # For business hours calculation, we need to collect the data and apply Python function
        if start_timestamp_key == timestamp_key:
            # Handle case where both keys are the same - use 0 duration
            df_collected = lf.collect()
            diff_values = [0.0] * len(df_collected)
        else:
            df_collected = lf.collect()
            diff_values = []
            
            for row in df_collected.iter_rows():
                start_idx = df_collected.columns.index(start_timestamp_key)
                end_idx = df_collected.columns.index(timestamp_key)
                
                diff = soj_time_business_hours_diff(
                    row[start_idx],
                    row[end_idx],
                    business_hours_slots,
                    workcalendar,
                )
                diff_values.append(diff)
        
        df_with_diff = df_collected.with_columns(
            pl.Series("__diff", diff_values)
        ).lazy()
    else:
        # Regular time difference calculation
        if start_timestamp_key == timestamp_key:
            # Handle case where both keys are the same - use 0 duration
            df_with_diff = lf.with_columns(pl.lit(0.0).alias("__diff"))
        else:
            df_with_diff = lf.with_columns(
                (pl.col(timestamp_key) - pl.col(start_timestamp_key))
                .dt.total_seconds()
                .alias("__diff")
            )

    # Group by activity and apply aggregation measure
    if aggregation_measure == "median":
        result = df_with_diff.group_by(activity_key).agg(pl.col("__diff").median())
    elif aggregation_measure == "min":
        result = df_with_diff.group_by(activity_key).agg(pl.col("__diff").min())
    elif aggregation_measure == "max":
        result = df_with_diff.group_by(activity_key).agg(pl.col("__diff").max())
    elif aggregation_measure == "sum":
        result = df_with_diff.group_by(activity_key).agg(pl.col("__diff").sum())
    else:  # mean
        result = df_with_diff.group_by(activity_key).agg(pl.col("__diff").mean())

    result_df = result.collect()
    
    # Convert to dictionary
    ret_dict = {}
    for row in result_df.iter_rows():
        activity = row[0]
        value = row[1]
        ret_dict[activity] = float(value) if value is not None else 0.0

    return ret_dict


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.util import pandas_utils


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Adds for each object an one-hot-encoding of the activities performed in its lifecycle

    Parameters
    -----------------
    ocel
        OCEL
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    data
        Values of the added features
    feature_names
        Names of the added features
    """
    if parameters is None:
        parameters = {}

    ordered_objects = (
        parameters["ordered_objects"]
        if "ordered_objects" in parameters
        else ocel.objects[ocel.object_id_column].to_numpy()
    )

    activities = pandas_utils.format_unique(
        ocel.events[ocel.event_activity].unique()
    )
    lifecycle = (
        ocel.relations.groupby(ocel.object_id_column)[ocel.event_activity]
        .agg(list)
        .to_dict()
    )

    data = []
    feature_names = ["@@ocel_lif_activity_" + str(x) for x in activities]

    for obj in ordered_objects:
        data.append([])
        if obj in lifecycle:
            lif = lifecycle[obj]
        else:
            lif = []
        for act in activities:
            data[-1].append(float(len(list(x for x in lif if x == act))))

    return data, feature_names



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.petri_net.obj import Marking


class DataMarking(Marking):
    def __init__(self, marking=None):
        Marking.__init__(self, marking)
        self.data_dict = {}

    def __repr__(self):
        # return str([str(p.name) + ":" + str(self.get(p)) for p in self.keys()])
        # The previous representation had a bug, it took into account the order
        # of the places with tokens
        return (
            str(
                [
                    str(p.name) + ":" + str(self.get(p))
                    for p in sorted(list(self.keys()), key=lambda x: x.name)
                ]
            )
            + " "
            + str(self.data_dict)
        )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Tuple


class Parameters(Enum):
    INCLUDE_TIMESTAMPS = "include_timestamps"


def apply(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Tuple[Tuple[Any, ...], Tuple[Any, ...]]:
    """
    Provides a tuple-based description of the provided object-centric event log

    Parameters
    --------------
    ocel
        Object-centric event log
    parameters
        Possible parameters of the algorithm, including:
        - Parameters.INCLUDE_TIMESTAMPS => include the timestamps (or not) in the representation

    Returns
    -------------
    Tuple[Tuple[Any, ...], Tuple[Any, ...]]
        Tuple-based representation of events and objects
    """
    if parameters is None:
        parameters = {}

    include_timestamps = exec_utils.get_param_value(
        Parameters.INCLUDE_TIMESTAMPS, parameters, True
    )

    object_ots = ocel.objects[
        [ocel.object_id_column, ocel.object_type_column]
    ].to_dict("records")
    object_ots = {
        x[ocel.object_id_column]: x[ocel.object_type_column]
        for x in object_ots
    }
    events = ocel.events.sort_values(
        [ocel.event_timestamp, ocel.event_activity, ocel.event_id_column]
    ).to_dict("records")
    objects = ocel.objects.sort_values(ocel.object_id_column).to_dict(
        "records"
    )
    relations = ocel.relations.sort_values(
        [
            ocel.event_timestamp,
            ocel.event_activity,
            ocel.object_id_column,
            ocel.event_id_column,
        ]
    )
    tdf = relations.groupby(ocel.object_id_column)[ocel.event_timestamp]
    objects_start = tdf.first().to_dict()
    objects_end = tdf.last().to_dict()
    objects_lifecycle = {
        x: objects_end[x].timestamp() - objects_start[x].timestamp()
        for x in objects_start
    }

    relations = (
        relations.groupby(ocel.event_id_column)[ocel.object_id_column]
        .agg(list)
        .to_dict()
    )

    event_descriptions = []

    for ev in events:
        related_objects = tuple(relations[ev[ocel.event_id_column]])
        event_tuple = (ev[ocel.event_activity],) + related_objects
        if include_timestamps:
            event_tuple = event_tuple + (ev[ocel.event_timestamp],)
        event_descriptions.append(event_tuple)

    object_descriptions = []

    for obj in objects:
        obj_id = obj[ocel.object_id_column]
        object_tuple = (obj_id, object_ots[obj_id])
        if include_timestamps:
            object_tuple = (
                object_tuple
                + (
                    objects_start[obj_id],
                    objects_end[obj_id],
                    objects_lifecycle[obj_id],
                )
            )
        object_descriptions.append(object_tuple)

    return tuple(event_descriptions), tuple(object_descriptions)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.causal.variants import alpha, heuristic
from enum import Enum
from pm4py.util import exec_utils
from typing import Dict, Tuple


class Variants(Enum):
    CAUSAL_ALPHA = alpha
    CAUSAL_HEURISTIC = heuristic


CAUSAL_ALPHA = Variants.CAUSAL_ALPHA
CAUSAL_HEURISTIC = Variants.CAUSAL_HEURISTIC

VERSIONS = {CAUSAL_ALPHA, CAUSAL_HEURISTIC}


def apply(
    dfg: Dict[Tuple[str, str], int], variant=CAUSAL_ALPHA
) -> Dict[Tuple[str, str], int]:
    """
    Computes the causal relation on the basis of a given directly follows graph.

    Parameters
    -----------
    dfg
        Directly follows graph
    variant
        Variant of the algorithm to use:
            - Variants.CAUSAL_ALPHA
            - Variants.CAUSAL_HEURISTIC

    Returns
    -----------
    causal relations
        dict
    """
    return exec_utils.get_variant(variant).apply(dfg)
